#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay_basic.h>
#include <stdint.h>


#define BIT0 (1<<0)
#define BIT1 (1<<1)
#define BIT2 (1<<2)
#define BIT3 (1<<3)
#define BIT4 (1<<4)
#define BIT5 (1<<5)
#define BIT6 (1<<6)
#define BIT7 (1<<7)

#define BUTTON BIT2

#define TOP          BIT0
#define TOP_LEFT     BIT7
#define TOP_RIGHT    BIT1
#define MIDDLE       BIT6
#define BOTTOM_LEFT  BIT2
#define BOTTOM_RIGHT BIT4
#define BOTTOM       BIT3
#define DOT          BIT5

#define OFF 10
#define BLANK 255

const uint8_t digit_select[2] = { BIT0, BIT1 };

static const uint8_t PROGMEM random_bits[256] =
{
69,   4,  78,  51,  36,  52,  76,  56,  40,  50,  34,  28,  44,  15,  66,  90, 
37,  96,  84,  98,  44,  65,  19,   5,  66,  44,  26,  67,  62,   6,  57,  95, 
43,  39,  77,  23,  53,  51,  97,   5,  51,  78,  65,  40,  10,  15,  93,  64, 
11,  45,  41,  13,  76,  42,  57,  34,  19,   1,  20,  33,  11,  35,  44,  56, 
69,  47,  46,  97,  17,  26,  70,  58,  15,   3,  46,   8,  92,  66,  17,  62, 
29,  55,  99,  88,  58,  54,  15,  70,  81,  65,  46,  90,  16,  78,  85,  56, 
40,   0,  98,  25,  50,  56,  26,  76,  86,  46,  37,  20,  96,  58,   6,   0, 
78,  67,  18,  93,  11,  85,  70,  51,  66,  38,  52,   7,   5,  22,  45,   0, 
58,  57,  18,  89,   9,  72,  35,  67,  75,  80,  16,  62,  42,  59,  54,  44, 
 4,  36,  98,  24,  78,  35,  60,   8,  30,  86,  63,  30,  66,  15,  96,  25, 
32,  89,  16,  86,  91,  28,  67,  59,  98,  84,  42,  68,   9,  98,  47,  68, 
85,  30,  98,  42,  22,  74,  11,  76,  19,  65,  51,  59,  94,  49,  35,  53, 
63,  88,  85,  23,   7,  81,  14,  42,  44,  48,  92,  81,  33,  69,  97,  79, 
94,  77,  87,   4,  56,  12,  83,  65,  13,  69,   4,  11,   0,  28,  74,  16, 
27,  75,  17,  31,  30,  24,  57,  69,  82,  65,  12,  16,  32,  47,  75,  21, 
70,  90,  90,  32,  11,   4,  78,  58,  42,  45,  84,  31,  48,  73,  59,   9, 
};

const uint8_t div_by_5[50] =
{
	0,0,0,0,0,
	1,1,1,1,1,
	2,2,2,2,2,
	3,3,3,3,3,
	4,4,4,4,4,
	5,5,5,5,5,
	6,6,6,6,6,
	7,7,7,7,7,
	8,8,8,8,8,
	9,9,9,9,9,
};

const uint8_t number_segments[12] =
{
	TOP|TOP_LEFT|TOP_RIGHT|BOTTOM_LEFT|BOTTOM_RIGHT|BOTTOM,        /* 0 */
	TOP_RIGHT|BOTTOM_RIGHT,                                        /* 1 */
	TOP|TOP_RIGHT|MIDDLE|BOTTOM_LEFT|BOTTOM,                       /* 2 */
	TOP|TOP_RIGHT|MIDDLE|BOTTOM_RIGHT|BOTTOM,                      /* 3 */
	TOP_LEFT|TOP_RIGHT|MIDDLE|BOTTOM_RIGHT,                        /* 4 */
	TOP|TOP_LEFT|MIDDLE|BOTTOM_RIGHT|BOTTOM,                       /* 5 */
	TOP|TOP_LEFT|MIDDLE|BOTTOM_LEFT|BOTTOM_RIGHT|BOTTOM,           /* 6 */
	TOP|TOP_RIGHT|BOTTOM_RIGHT,                                    /* 7 */
	TOP|TOP_LEFT|TOP_RIGHT|MIDDLE|BOTTOM_LEFT|BOTTOM_RIGHT|BOTTOM, /* 8 */
	TOP|TOP_LEFT|TOP_RIGHT|MIDDLE|BOTTOM_RIGHT|BOTTOM,             /* 9 */
	0,                                                             /* off */
	DOT,
};

void run(void);
void init(void);

int main(void)
{
	init();
	run();
}

void display_digit(uint8_t d)
{
	PORTA = number_segments[d];
	__builtin_avr_delay_cycles(1200);
}

void display_off(void)
{
	PORTA = number_segments[OFF];
}

void select_digit(uint8_t which)
{
	PORTB = digit_select[which];
}

void display(uint8_t n, uint8_t iters)
{
	uint8_t i, d0, d1;

	if (n < 100)
	{
		d0 = div_by_5[n>>1];
		d1 = n - ( (d0<<3) + (d0<<1) );
		if (d0 == 0)
			d0 = OFF;
	}
	else
		d0 = d1 = OFF;

	for (i=0; i<iters; i++)
	{
		display_off();
		select_digit(0);
		__builtin_avr_delay_cycles(150);
		display_digit(d0);
		display_off();
		select_digit(1);
		__builtin_avr_delay_cycles(150);
		display_digit(d1);
	}
}

void button_push(void)
{
	uint8_t i, j;

	for (i=0; i<15; i++)
	{
		for (j=0; j<5; j++)
			display(42, 255);

		for (j=0; j<5; j++)
			display(BLANK, 255);
	}
}

void init(void)
{
	GIMSK |= 0x40;
	SREG |= 0x80;
	DDRA = 0xff;
	DDRB = 0x0f & ~BUTTON;
}

ISR(INT0_vect)
{
	button_push();
}

void run(void)
{
	uint8_t n, fade, cur, next;

	next = pgm_read_byte(&random_bits[n]);
	for ( ;; )
	for (n=0;n<255;n++)
	{
		cur = next;
		next = pgm_read_byte(&random_bits[n]);

		for (fade=0; fade<56; fade++)
			display(cur, 63);

		/* Fade transition loop */
		for (fade=0; fade<40; fade++)
		{
			display(next, fade);
			display(cur, 40-fade);
		}
	}
}

